﻿    using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.U2D;

public class Cell : MonoBehaviour, ISpaceForGas
{
    // Types and quantities of each gas in this cell
    public Dictionary<Gas.GasType, float> cellGas = new Dictionary<Gas.GasType, float>();

    // Total gas currently in this cell, all types of gas
    public float totalGas {get; set;} = 0.0f;

    // Coordinates within a room. Top left cell has position [0, 0],
    // or perhaps [0 + cell_spacing, 0 + cell_spacing]
    public float pos_x;
    public float pos_y;

    // Cells neighboring this cell
    public List<ISpaceForGas> neighbors;
    public Cell north_cell;
    public Cell east_cell;
    public Cell south_cell;
    public Cell west_cell;

    public Color tempColor;
    public System.Random ran = new System.Random();

    // How fast gas moves to neighbors. rate per second.
    public float gasMoveRate = 20000.0f;
    public float fireDecayRate = 100.0f;
    public float gasBurnRate = 64000.0f;

    // How often gas moves randomly instead of preferring moving to
    // neighbor with least total gas. [0.0, 1.0] where 1.0 is never random
    public float gasRandomFactor = 0.8f;

    public void initGas() {
        cellGas.Add(Gas.GasType.GasAir, 0.0f);
        cellGas.Add(Gas.GasType.GasInflammable, 0.0f);
        cellGas.Add(Gas.GasType.GasPoison, 0.0f);
        cellGas.Add(Gas.GasType.GasFire, 0.0f);
    }

    // Randomly shuffle the order of items in a list
    void ShuffleNeighbors() {
        int n = neighbors.Count;
        while (n > 1) {
            n--;
            int k = ran.Next(n + 1);
            ISpaceForGas value = neighbors[k];
            neighbors[k] = neighbors[n];
            neighbors[n] = value;
        }
    }

    static int SortByTotalGas(ISpaceForGas a, ISpaceForGas b)
    {
        return a.totalGas.CompareTo(b.totalGas);
    }

    public void addGas(Gas.GasType gt, float quantity) {

        var gasName = Gas.getGasName(gt);

        cellGas[gt] += quantity;
        if (cellGas[gt] < 0.0f) {
            cellGas[gt] = 0.0f;
        } else if (cellGas[gt] > 100.0f) {
            cellGas[gt] = 100.0f;
        }

        foreach (var g in cellGas) {
            totalGas += g.Value;
        }
    }

    // Start is called before the first frame update
    void Start() {
    }

    // Update is called once per frame
    void Update()
    {
        // See if the room has an output cell where gas must head to
        // TODO: assign
        Cell output = null;
        if (output != null) {
            Cell target = null;
            float diff_x = pos_x - output.pos_x;
            float diff_y = pos_y - output.pos_y;

            bool use_topdown = System.Math.Abs(diff_x) > System.Math.Abs(diff_y);

            // TODO: When it's available, use room's property cell_spacing here
            float cell_spacing = 0.1f;
            if (use_topdown && diff_x <= cell_spacing) {
                use_topdown = false;
            }

            // See if the route is more to up/down or left/right axis
            // TODO: Check for scenario where cell is next to wall
            if (use_topdown == false) {
                if (diff_x > 0.0f) {
                    // left neighbor
                } else {
                    // right neighbor
                }
            } else {
                if (diff_y > 0.0f) {
                    // bottom neighbor
                } else {
                    // top neighbor
                }
            }

            List<Gas.GasType> keyList = new List<Gas.GasType>(this.cellGas.Keys);

            if (target == null) {
                // If no target was found, begin fading all gas in this cell
                foreach (var gt in keyList) {
                    cellGas[gt] = (float)System.Math.Floor(cellGas[gt] / 2.0f);
                    if (cellGas[gt] < 2.0f) {
                        cellGas[gt] = 0.0f;
                    }
                }
            } else {
                // otherwise give the neighbor half of the gas
                foreach (var gt in keyList) {
                    float to_remove = (float)System.Math.Floor(cellGas[gt] / 2.0f);
                    target.cellGas[gt] += to_remove;
                    cellGas[gt] -= to_remove;

                    if (cellGas[gt] < 2.0f) {
                        cellGas[gt] = 0.0f;
                    }

                    // TODO: Instead of clamping gas, move it to next available tile.
                    // Currently gas will coalesce into one tile and will get destroyed.
                    if (target.cellGas[gt] > 100.0f) {
                        target.cellGas[gt] = 100.0f;
                    }
                }
            }
        }

        // Spread gas to neighbors
        if (neighbors != null && neighbors.Count > 0) {

            // get random direction to spread gas
            // try until we find non-null or run out of neighbors
            ISpaceForGas n = null;

            // Prefer sorting neighbors by how little gas they have, but
            // occasionally arrange them randomly.
            if (Random.Range(0.0f, 1.0f) > gasRandomFactor) {
                neighbors.Sort(SortByTotalGas);
            } else {
                ShuffleNeighbors();
            }

            int count = 0;
            while (count < neighbors.Count) {
                if (neighbors[count] != null) {
                    n = neighbors[count];
                    break;
                }
                count++;
            }

            handleCellNeighbor(n);
        }

        handleGasMechanics();
        setGasColor();
    }

    void handleGasMechanics() {

        Cell this_c = (Cell)this;
        List<Gas.GasType> keyList = new List<Gas.GasType>(this_c.cellGas.Keys);

        // If the call has fire in it, burn some away while consuming other gasses
        if (cellGas[Gas.GasType.GasFire] > 0.0f) {
            float to_remove = fireDecayRate * Time.deltaTime;
            float fire_remove = System.Math.Min(cellGas[Gas.GasType.GasFire], to_remove);

            // Decay all combustible gasses
            //foreach (var gt in keyList) {
            //    if (Gas.isCombustible(gt)) {
            //        addGas(gt, -fire_remove);
            //    }
            //}

            // Ignite neighbors with combustible gasses
            foreach (var c in neighbors) {
                if (c is GasInput) {
                    continue;
                }

                Cell cell_c = (Cell)c;

                foreach (var gt in keyList) {
                    if (Gas.isCombustible(gt)) {
                        if (this_c.cellGas[gt] > 0.0) {
                            float n_remove = System.Math.Min(this_c.cellGas[gt], to_remove);
                            cell_c.addGas(gt, -n_remove);
                            cell_c.addGas(Gas.GasType.GasFire, n_remove);
                        }
                    }
                }
            }
        }

        if (cellGas[Gas.GasType.GasAir] > 20.0f && cellGas[Gas.GasType.GasInflammable] > 20.0f) {
            // Form fire by combining the two gasses. This then starts a chain reaction if
            // combustible gasses are present in the neighbors.
            float to_remove = gasBurnRate * Time.deltaTime;
            float air_remove = System.Math.Min(cellGas[Gas.GasType.GasAir], to_remove);
            float inf_remove = System.Math.Min(cellGas[Gas.GasType.GasInflammable], to_remove);
            addGas(Gas.GasType.GasAir, -air_remove);
            addGas(Gas.GasType.GasInflammable, -inf_remove);
            addGas(Gas.GasType.GasFire, air_remove + inf_remove);
            Debug.LogFormat("FIRE");
        }

        if (cellGas[Gas.GasType.GasAir] > 20.0f && cellGas[Gas.GasType.GasPoison] > 20.0f) {
            // Air slowly transforms into poison gas
        }

        if (cellGas[Gas.GasType.GasInflammable] > 20.0f && cellGas[Gas.GasType.GasPoison] > 20.0f) {
            // A new type of gas is formed?
        }
    }

    public void handleCellNeighbor(ISpaceForGas n) {
        List<Gas.GasType> keyList = new List<Gas.GasType>(this.cellGas.Keys);
        float transfer = gasMoveRate * Time.deltaTime;

        foreach (var gt in keyList) {
            if (cellGas[gt] > 0.5) {

                // Could use some sort of curve here but maths is hard :(
                if (cellGas[gt] > 80) {
                    transfer *= 100;
                } else if (cellGas[gt] > 60) {
                    transfer *= 50;
                } else if (cellGas[gt] > 40) {
                    transfer *= 20;
                } else if (cellGas[gt] > 20) {
                    transfer *= 10;
                }

                float to_move = System.Math.Min(cellGas[gt], transfer);

                if (n is GasInput) {
                    GasInput gi_n = (GasInput)n;
                    if (gi_n.isOpen) {
                        gi_n.addGas(gt, to_move);
                        addGas(gt, -to_move);
                    }
                } else {
                    Cell cell_n = (Cell)n;
                    // Only transfer gas if the neighbor has less and some gas would remain behind
                    // in this cell after the transfer
                    if (cell_n.cellGas[gt] < cellGas[gt] && cellGas[gt] - to_move > 0.0) {
                        cell_n.addGas(gt, to_move);
                        addGas(gt, -to_move);
                    }
                }
            }
        }
    }

    float lerp(float a, float b, float alpha) {
        return a * (1.0f - alpha) + alpha * b;
    }

    void setGasColor() {
        tempColor.a = 1.0f;
        if (cellGas[Gas.GasType.GasFire] > 0.0f) {
            tempColor.r = lerp(0.3f, 1.0f, cellGas[Gas.GasType.GasFire] / 100.0f);
            tempColor.g = lerp(0.3f, 1.0f, cellGas[Gas.GasType.GasFire] / 100.0f);
            tempColor.b = 0.0f;
        } else if (cellGas[Gas.GasType.GasInflammable] > 0.0f) {
            tempColor.r = lerp(0.3f, 1.0f, cellGas[Gas.GasType.GasInflammable] / 100.0f);
            tempColor.g = 0.0f;
            tempColor.b = 0.0f;
        } else if (cellGas[Gas.GasType.GasPoison] > 0.0f) {
            tempColor.r = lerp(0.3f, 1.0f, cellGas[Gas.GasType.GasPoison] / 100.0f);
            tempColor.g = 0.0f;
            tempColor.b = lerp(0.3f, 1.0f, cellGas[Gas.GasType.GasPoison] / 100.0f);
        } else if (cellGas[Gas.GasType.GasAir] > 0.0f) {
            tempColor.r = 0.0f;
            tempColor.g = 0.0f;
            tempColor.b = lerp(0.3f, 1.0f, cellGas[Gas.GasType.GasAir] / 100.0f);
        } else {
            tempColor.r = 0.0f;
            tempColor.g = 0.0f;
            tempColor.b = 0.0f;
        }
            this.GetComponent<SpriteShapeRenderer>().color = tempColor;
        }
}
