﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.U2D;

[ExecuteInEditMode]
public class Room : MonoBehaviour
{
    public float cellScaler = 0.5f;
    public GameObject cellPrefab;
    private GameObject[,] roomCells;
    private GasInput[] gasInputs;

    int xCells;
    int yCells;

    float xFactor;
    float yFactor;

    public bool isVacuum;
    public Gas.GasType startingGasType;
    public float startingGasQuantity;

    public bool generateInEditor = false;

    private float prevScale = 0;

    // Start is called before the first frame update
    void Start()
    {
        if (!Application.isPlaying && !generateInEditor) {
            return;
        }
        // Init if init hasn't happened in editor yet
        if(roomCells == null || roomCells.Length == 0) {
            prevScale = cellScaler;
            Init();
        }
    }

    public void addGasToCell(int xPos, int yPos, Gas.GasType gasType, float gasQuantity) {
        roomCells[xPos, yPos].GetComponent<Cell>().addGas(gasType, gasQuantity);
    }

    void KillAllChildrenExceptIO() {
        int childCount = this.transform.childCount;
        Debug.Log("Killing children. Amount of children: " + childCount);

        var children = new List<GameObject>();
        foreach(Transform child in transform) {
            if (!child.CompareTag("GasIO")) {
                children.Add(child.gameObject);
            }
        }

        int count = 0;
        children.ForEach(child => {
            if(Application.isEditor) {
                DestroyImmediate(child);
            } else {
                Destroy(child);
            }
            count++;
        });
        Debug.LogFormat("Killed {0} children", count);
    }

    void initGasInputs() {
        gasInputs = transform.GetComponentsInChildren<GasInput>();
    }

    void Init() {
        // Clear the state if editor has initialized this before
        KillAllChildrenExceptIO();
        initGasInputs();

        SpriteShapeRenderer ownRenderer = this.GetComponent<SpriteShapeRenderer>();
        Vector3 extents = ownRenderer.bounds.extents/2.0f;
        Debug.Log("Room bounds: " + ownRenderer.bounds);

        // Hide own renderer. It's just for dev purposes.
        ownRenderer.enabled = false;

        xCells = Mathf.RoundToInt(extents.x / cellScaler);
        yCells = Mathf.RoundToInt(extents.y / cellScaler);
        float xPositionStep = 2.0f/xCells;
        float yPositionStep = 2.0f/yCells;
        xFactor = 2.0f/xCells;
        yFactor = 2.0f/yCells;

        Debug.LogFormat("Cells to be generated: X={0}, Y={1}", xCells, yCells);
        if (xCells < 0.0f || yCells < 0.0f) {
            return;
        }

        roomCells = new GameObject[xCells, yCells];
        for(int i = 0; i < xCells; i++) {
            for(int j = 0; j < yCells; j++) {
                roomCells[i, j] = InstantiateCell(i, j, new Vector3(-2.0f + xFactor + xPositionStep*2 * i, 2.0f -yFactor - yPositionStep*2 * j, 0));
            }
        }
        PopulateNeighbors();
    }

    private GameObject InstantiateCell(int xPos, int yPos, Vector3 position) {
        GameObject newCell = Instantiate(cellPrefab, new Vector3(0,0,0), Quaternion.identity);
        newCell.transform.parent = this.transform;
        newCell.transform.localPosition = position;
        newCell.transform.localScale = new Vector3(xFactor, yFactor, 1.0f);
        newCell.GetComponent<SpriteShapeRenderer>().color = new Color(Random.Range(0.0f, 1.0f), Random.Range(0.0f, 1.0f), Random.Range(0.0f, 1.0f));
        Cell cell = newCell.GetComponent<Cell>();
        cell.pos_x = xPos;
        cell.pos_y = yPos;
        cell.initGas();

        if (!isVacuum) {
            cell.addGas(startingGasType, startingGasQuantity);
        }
        return newCell;
    }

    private void PopulateNeighbors() {
        for(int i = 0; i < xCells; i++) {
            for(int j = 0; j < yCells; j++) {
                var neighbors = new List<ISpaceForGas>();

                // North
                if (j == 0) {
                    roomCells[i, j].GetComponent<Cell>().north_cell = null;
                } else {
                    neighbors.Add(roomCells[i, j - 1].GetComponent<Cell>());
                    roomCells[i, j].GetComponent<Cell>().north_cell = roomCells[i, j - 1].GetComponent<Cell>();
                }
                // East
                if (i == xCells - 1) {
                    roomCells[i, j].GetComponent<Cell>().east_cell = null;
                } else {
                    neighbors.Add(roomCells[i + 1, j].GetComponent<Cell>());
                    roomCells[i, j].GetComponent<Cell>().east_cell = roomCells[i + 1, j].GetComponent<Cell>();
                }
                // South
                if (j == yCells - 1) {
                    roomCells[i, j].GetComponent<Cell>().south_cell = null;
                } else {
                    neighbors.Add(roomCells[i, j + 1].GetComponent<Cell>());
                    roomCells[i, j].GetComponent<Cell>().south_cell = roomCells[i, j + 1].GetComponent<Cell>();
                }
                // West
                if (i == 0) {
                    roomCells[i, j].GetComponent<Cell>().west_cell = null;
                } else {
                    neighbors.Add(roomCells[i - 1, j].GetComponent<Cell>());
                    roomCells[i, j].GetComponent<Cell>().west_cell = roomCells[i - 1, j].GetComponent<Cell>();
                }

                foreach(var input in gasInputs) {
                    if (input.xPos == i && input.yPos == j) {
                        neighbors.Add(input);
                    }
                }

                roomCells[i, j].GetComponent<Cell>().neighbors = neighbors;
            }
        }
    }

    // Update is called once per frame
    void Update()
    {
        if (prevScale != cellScaler) {
            prevScale = cellScaler;
            if (Application.isEditor && !generateInEditor) {
                return;
            }
            Init();
        }
    }
}
